from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from easybits_module_interface.models.base import BaseDataClass
from easybits_module_interface.models.enum import MessageTypes, SystemMessageEvents


@dataclass
class ContextProvider(BaseDataClass):
    """
    Configuration for a context provider
    """
    #: name of resulting key in context
    key: str = ""
    #: name of the context provider to use
    provider_name: str = ""


@dataclass
class AvailableModule(BaseDataClass):
    """
    Module config template
    """
    #: Internal ID
    id: int
    #: verbose identifier of module
    name: str
    #: description of module
    description: str
    #: indicates that the module supports reengagement
    supports_reengagement: bool = False
    #: Layout of configuration values
    configuration_layout: Optional[Dict] = None
    #: List of supported context providers by the module
    supported_context_providers: Optional[List] = None
    #: Indicates if the module is allowed to be used
    #:   without any pre-selection process
    is_passive: bool = False
    # The following attributes are commented out, as they
    # are only used in the client applications.
    #: Indicates if the module is available to be used
    #is_published: bool = True
    #: Indicates if the module is visible for users
    #is_visible: bool = True
    #: Indicates if the module is the default module
    #    that is used during the default module prediction
    #    phase of the parser.
    is_default: bool = False
    #: String of comma separated intentions that the module supports
    intentions: Optional[str] = None


@dataclass
class BotModule(BaseDataClass):
    """
    Configuration for a bot specific module.
    """
    #: internal identifier of bot module
    id: int
    #: instance of :py:obj:`~.AvailableModule`
    module: AvailableModule = None
    #: indicates that the module is active
    is_active: bool = False
    #: indicates that the module uses reengagement
    has_reengagement: bool = False
    #: list of instances of :py:obj:`~.ContextProvider` \
    #  that are used to extract context from incoming messages
    context_providers: List[ContextProvider] = field(default_factory=list)
    #: configuration values for the module
    configuration_values: Dict = field(default_factory=dict)
    #: configuration values for the parser
    parser_configuration: Dict = field(default_factory=dict)
    #: desired reengagement message
    reengagement_message: str = None
    #: Timeout in minutes for reengagement message
    reengagement_timeout_minutes: int = None

    @property
    def name(self):
        return self.module.name

    @property
    def intentions(self):
        return self.module.intentions

    @classmethod
    def from_dict(cls, **kwargs):
        """
        Creates an instance from given kwargs

        :param kwargs: object as kwargs
        :return: :py:obj:`~.BotModule` instance
        """
        module = kwargs.pop('module', None)
        context_providers = kwargs.pop('context_providers', [])

        return super().from_dict(
            **kwargs,
            module=AvailableModule.from_dict(**module) if module else None,
            context_providers=[ContextProvider.from_dict(**c) for c in context_providers]
        )


@dataclass
class ConfigureModuleContext(BaseDataClass):
    """
    Integration config template
    """
    id: str = None
    #: name of the integration
    name: str = None
    #: list of instances of :py:obj:`~.ContextProvider`
    context_providers: List[ContextProvider] = field(default_factory=list)
    #: indicates that the integration is ready to consume messages
    is_active: bool = False
    #: indicates that the integration supports reengagement
    re_engagement: bool = False

    #: layout of configuration values
    configuration_layout: Optional[Dict] = None
    #: indicates that the module supports reengagement
    supports_reengagement: bool = False
    #: description of module
    description: str = None






import asyncio
import logging
import sentry_sdk
import os
from aio_pika import connect_robust


async def bootstrap_module_class(module_implementation, **config) -> None:
    """
    Main entry point for module implementations, aka application loop.

    :param module_implementation: ModuleInterface implementation (as class)
    :param config: Configuration dictionary
    :return: None

    ================  =============  =============
    Key               Type           Description
    ================  =============  =============
    logger            dict           Logger configuration
    logger.level      str            Logger level
    logger.path       str            Path to logger configuration file
    logger.name       str            Name of logger
    ================= =============  =============
    broker            dict           Broker configuration
    broker.uri        str            Broker uri
    """
    logger_config = config.get('logger', None)
    broker_config = config.get('broker', None)

    assert 'uri' in broker_config, 'broker uri required!'

    logger_name = None
    if logger_config is not None:
        logging.basicConfig()
        if 'level' in logger_config:
            logging.getLogger().setLevel(logger_config['level'])

        if 'path' in logger_config:
            # set up logger from file
            logging.config.fileConfig(logger_config['path'])

        # create logger
        logger_name = logger_config.get('name', 'module_interface')
        _ = logging.getLogger(name=logger_name)

    connection = await connect_robust(broker_config['uri'])
    module = module_implementation(connection)
    module.init_logger(logger_name)
    return module


async def main(module_implementation, **config) -> None:
    """
    Main entry point for module implementations, aka application loop.

    :param module_implementation: ModuleInterface implementation (as class)
    :param config: Configuration dictionary
    :return: None

    ================  =============  =============
    Key               Type           Description
    ================  =============  =============
    logger            dict           Logger configuration
    logger.level      str            Logger level
    logger.path       str            Path to logger configuration file
    logger.name       str            Name of logger
    ================= =============  =============
    broker            dict           Broker configuration
    broker.uri        str            Broker uri
    """
    module = await bootstrap_module_class(module_implementation, **config)
    try:
        await module.run()
    except Exception as e:
        module.logger.error(e)
    finally:
        module.logger.info("MAIN Shutting down")
        await module.shutdown()


async def multitask_main(loop, module_cls, logger_name, logger_config_path, tasks):  # pragma: no cover
    """
    Main entry point for module implementations that requires multiple threads.

    This method starts all tasks and shuts down the application gracefully.

    :param loop: Your current asyncio loop
    :param module_implementation: ModuleInterface implementation (as class)
    :param logger_name: The name of your logger/application
    :param tasks: A list of strings containing method names of the `module_implementation`
                that will be executed in parallel
    """
    module = await bootstrap_module_class(
        module_cls,
        logger={
            "path": logger_config_path,
            "level": logging.DEBUG,
            "name": logger_name
        },
        broker={'uri': os.environ.get('RMQ_REMOTE_URL')}
    )

    task_map = []
    for t in tasks:
        task = loop.create_task(getattr(module, t)())
        task.add_done_callback(lambda x: module.logger.debug(f'Future: {x}'))
        task_map.append(task)

    try:
        await asyncio.gather(*task_map)
    except asyncio.CancelledError:
        module.logger.error("Main loop cancelled")
        await module.shutdown()


def setup_application():  # pragma: no cover
    VERSION = '0000000'
    if os.path.exists(os.path.join(os.path.dirname(__file__), '../version.txt')):
        with open(os.path.join(os.path.dirname(__file__), '../version.txt'), 'r') as f:
            VERSION = f.read().strip()

    sentry_sdk.init(
        dsn=os.environ.get('SENTRY_DSN', None),
        environment=os.environ.get('SENTRY_ENVIRONMENT', None),
        release=VERSION,
    )




from ._internal import ModuleInterface  # noqa
from ._app_helpers import bootstrap_module_class, main, multitask_main, setup_application  # noqa
from ._version import __version__  # noqa
from .task_execution.task_executor import TaskExecutor
from .task_execution.models import GenerationRequest
from .state_machine import (IllegalTransitionError, StateMachine,  # noqa
                            Transition)
from .logger import Formatter
from . import models

__all__ = [
    '__version__',
    'main',
    'multitask_main',
    'models',
    'bootstrap_module_class',
    'setup_application',
    'ModuleInterface',
    'TaskExecutor',
    'GenerationRequest',
    'StateMachine',
    'Transition',
    'IllegalTransitionError',
    'Formatter',
]



from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from easybits_module_interface.models.base import DynamicFieldsDataClass, BaseDataClass
from easybits_module_interface.models.enum import MessageTypes, SystemMessageEvents
from easybits_module_interface.models.module import AvailableModule


@dataclass
class MessageMeta(DynamicFieldsDataClass):
    """
    Meta data for messages.

    This class is dynamic, meaning using the from_dict method
    will create a new dataclass with the fields provided in the kwargs.

    This is useful to add more meta data to the message without
    changing the class definition.
    """
    #: type of message
    type: MessageTypes
    #: event type of message
    event: Optional[str] = None
    #: internal ID of communication channel
    communicationChannelId: Optional[str] = None
    #: type identifier of the controller used for communication
    controller: Optional[str] = None
    #: ID of the user can be used for tracking/state management
    internalUserId: Optional[str] = None
    #: ID of the bot used for communication
    botId: Optional[str] = None

    #: [Optional] configured language of bot used inside modules to generate responses
    language: Optional[str] = None
    #: [Optional] internal tracking ID of the message
    messageId: Optional[str] = None


@dataclass
class UserMessage(BaseDataClass):
    """
    Object to describe message content aka a UserMessage
    """
    #: content of incoming message
    incoming: Optional[str] = None
    #: content of incoming file
    incoming_file: Optional[str] = None
    #: content of outgoing message
    outgoing: Optional[str] = None


@dataclass
class Message(BaseDataClass):
    #: message meta data
    meta: MessageMeta
    #: extracted context from incoming message
    context: Optional[Union[Dict, 'ConfigureModuleContext']] = None
    #: message content
    message: Optional[UserMessage] = None
    #: configuration values forwarded to the module
    config: Optional[Dict] = None

    @classmethod
    def from_dict(cls, **kwargs):
        """
        Creates an instance from given kwargs

        :param kwargs: object as kwargs
        :return: :py:obj:`~.Message` instance
        """
        meta_data = kwargs.pop('meta', None)
        if meta_data is not None:
            meta_data = MessageMeta.from_dict(**meta_data)
        message = kwargs.pop('message', None)
        if message is not None:
            message = UserMessage.from_dict(**message)

        context = kwargs.pop('context', None)
        is_configure_module_message = (
            meta_data
            and meta_data.type == MessageTypes.SYSTEM_MESSAGE_PARSER.value
            and meta_data.event in [
                SystemMessageEvents.SAVE.value,
                SystemMessageEvents.DELETE.value,
            ]
        )
        if is_configure_module_message:
            context = AvailableModule.from_dict(**context)

        config = kwargs.pop('config', None)

        return super().from_dict(
            context=context,
            meta=meta_data,
            message=message,
            config=config,
        )




from enum import Enum


class MessageTypes(Enum):
    SYSTEM_MESSAGE_PARSER = 'system_message_parser'
    SYSTEM_MESSAGE_CONTROLLER = 'system_message_controller'
    SYSTEM_MESSAGE_DATASINK = 'system_message_datasink'
    INCOMING_MESSAGE = 'incoming_message'
    OUTGOING_MESSAGE = 'outgoing_message'
    OUTGOING_MESSAGE_IMAGE = 'outgoing_message_image'
    OUTGOING_MESSAGE_AUDIO = 'outgoing_message_audio'


class SystemMessageEvents(Enum):
    SAVE = 'save'
    DELETE = 'delete'
    CONFIGURE_PROXY_MODE = 'configure_proxy_mode'






from copy import deepcopy
from dataclasses import dataclass, field, fields, make_dataclass
from typing import Dict, List, Optional, Union


class BaseDataClass:
    """
    Base class for dataclasses with a from_dict method.
    """
    @classmethod
    def get_model_kwargs(cls, kwargs: Dict) -> Dict:
        """
        Getter for kwargs based on the fields of the class.

        :param kwargs: kwargs to filter
        :return: filtered kwargs
        """
        field_names = [field.name for field in fields(cls)]
        kwargs_out = deepcopy(kwargs)
        for f in kwargs.keys():
            if f not in field_names:
                kwargs_out.pop(f, None)
        return kwargs_out

    @classmethod
    def from_dict(cls, **kwargs):
        """
        Safe method to create an instance from given kwargs.

        :param kwargs: object as kwargs
        :return: instance of the class
        """
        return cls(**cls.get_model_kwargs(kwargs))


class DynamicFieldsDataClass(BaseDataClass):
    """
    Base class for dataclasses with dynamic fields.

    Example:
        >>> @dataclass
        ... class MyFoo(DynamicFieldsDataClass):
        ...     foo: str
        ...
        >>> obj = MyFoo.from_dict(foo='string', bar=1)
        >>> obj.foo
        'string'
        >>> obj.bar
        1
        >>> asdict(obj)
        {'foo': 'string', 'bar': 1}
        >>> MyFoo(foo='string', bar=1)
        AttributeError: 'MyFoo' object has no attribute 'bar'
    """
    @classmethod
    def from_dict(cls, **kwargs):
        """
        Creates a new dataclass with the fields provided in the kwargs.

        New fields are added to the class definition as Optional[str].

        :param kwargs: object as kwargs
        :return: new dataclass instance
        """
        field_names = [field.name for field in fields(cls)]
        new_fields = [(k, Optional[str], None) for k, v in kwargs.items() if k not in field_names]
        dynamic_cls = make_dataclass(cls.__name__, new_fields, bases=(cls,))
        return dynamic_cls(**kwargs)





import logging


class Formatter(logging.Formatter):  # pragma: no cover
    grey = "\x1b[38;20m"  # ] closing bracket for IDE
    yellow = "\x1b[33;20m"  # ] closing bracket for IDE
    red = "\x1b[31;20m"  # ] closing bracket for IDE
    blue = "\x1b[34;20m"  # ] closing bracket for IDE
    bold_red = "\x1b[31;1m"  # ] closing bracket for IDE
    reset = "\x1b[0m"  # ] closing bracket for IDE
    fmt = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: grey + fmt + reset,
        logging.INFO: blue + fmt + reset,
        logging.WARNING: yellow + fmt + reset,
        logging.ERROR: red + fmt + reset,
        logging.CRITICAL: bold_red + fmt + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)




import asyncio
import json
import sentry_sdk
import traceback
import logging
import logging.config
import os
import base64
import mimetypes
import requests
from dataclasses import asdict
from typing import Callable, Optional
from aio_pika import Message as PikaMessage, ExchangeType
from aio_pika.abc import AbstractIncomingMessage

from easybits_module_interface.models import Message


class ConnectionError(Exception):
    """
    Exception raised when connection to RabbitMQ fails.
    """
    pass


class ConnectionSetup:
    """
    Base class for classes that need to connect to RabbitMQ.

    Provides a :py:obj:`~connection` and :py:obj:`~channel` attribute.

    :param connection: Connection to RabbitMQ

    Example::
        >>> import aio_pika
        >>> connection = await aio_pika.connect("amqp://localhost/", loop=loop)
        >>> setup = ConnectionSetup(connection)
        >>> await setup.connect()
        >>> await setup.channel.declare_queue("test")
        >>> await setup.disconnect()

    """
    def __init__(self, connection):
        self.connection = connection
        self.channel = None
        self.logger = logging.getLogger(__name__)

    def set_logger(self, logger) -> None:
        """
        Set logger for class.

        :param logger: Logger to be used
        :return: None
        """
        self.logger = logger

    async def connect(self, channel_id: Optional[int] = None) -> None:
        """
        Establish a connection (channel) to RabbitMQ.

        :param channel_id: Channel id to use
        :return: None
        :raises ConnectionError: If connection not established
        """
        if self.connection is None or self.connection.is_closed:
            raise ConnectionError("Connection not established")
        self.channel = await self.connection.channel(channel_id)

    async def disconnect(self) -> None:
        """
        Disconnect from RabbitMQ.

        :return: None
        """
        if self.channel is not None and not self.channel.is_closed:
            await self.channel.close()


class Publisher(ConnectionSetup):
    """
    Class to publish messages to RabbitMQ.

    Note: When `exchange != ""` the publisher will declare the exchange if it does not exist.

    :param connection: Connection to RabbitMQ
    :param exchange: Name of exchange to publish messages to

    Example::
        >>> import aio_pika
        >>> from easybits_module_interface import Publisher
        >>> from easybits_module_interface.models import Message
        >>> connection = await aio_pika.connect("amqp://localhost/", loop=loop)
        >>> publisher = Publisher(connection, "test_exchange")
        >>> await publisher.publish(Message.from_dict(**message_dict))
        >>> await publisher.disconnect()
    """
    def __init__(self, connection, exchange):
        super().__init__(connection)
        self.exchange_name = exchange
        self.default_exchange = None

    async def connect(self):
        """
        Establish a connection to RabbitMQ.

        :return: None
        """
        await super().connect()
        if self.exchange_name == "":
            self.logger.warning("No exchange name provided, using default exchange")
            self.default_exchange = self.channel.default_exchange
            return
        # ensure default exchange exists
        self.default_exchange = await self.channel.declare_exchange(
            self.exchange_name, ExchangeType.FANOUT, durable=True
        )

    async def publish(self, message: Message, exchange: Optional[str] = None, queue: str = ""):
        """
        Publish a message to RabbitMQ.

        Internally, establishes a connection via a new channel, and publishes the message to the desired target.
        Converts :py:obj:`~message` to a dictionary, then to a JSON string and finally to a byte string.
        If :py:obj:`~exchange` is empty, the message will be published to the Publisher's default exchange.

        **Note**: <u>Please do not publish messages to the default exchange & default queue,
                  always provide a queue or exchange name.</u>

        Values for :py:obj:`~exchange`:
            - None: Use the default exchange of the module
            - "": Use the RMQ default exchange
            - "exchange_name": Use the exchange with the given name

        :param message: Message to be published
        :param exchange: Name of exchange to publish message to
        :param queue: Name of queue to publish message to
        :return: None
        """
        if self.channel is None or self.channel.is_closed:
            await self.connect()

        if exchange is None:
            ex = self.default_exchange
        elif exchange == "":
            ex = self.channel.default_exchange
        else:
            ex = await self.channel.get_exchange(exchange)

        msg_body = json.dumps(asdict(message)).encode('utf-8')
        pika_msg = PikaMessage(body=msg_body)
        await ex.publish(
            pika_msg,
            routing_key=queue
        )


class Consumer(ConnectionSetup):
    """
    Class to consume messages from RabbitMQ.

    :param connection: Connection to RabbitMQ

    Example::
        >>> import aio_pika
        >>> from easybits_module_interface import Consumer
        >>> from easybits_module_interface.models import Message
        >>> connection = await aio_pika.connect("amqp://localhost/", loop=loop)
        >>> consumer = Consumer(connection)
        >>> consumer.add_callback("test_queue", callback)
        >>> await consumer.run()
        >>> await consumer.disconnect()
    """
    def __init__(self, connection):
        super().__init__(connection)
        self.queue = None
        self.callback_map = {}

    @staticmethod
    def __callback_wrapper(fun):
        """
        Wraps input :py:obj:`~fun` (callbacks) so that signature matches :py:obj:`~pika`s consumption call.
        Internally builds :py:obj:`~models.Message` from incoming message body, and passes it to :py:obj:`~fun`.

        This way the callback can use Message class to interact with the message instead of handling a dictionary.

        :param fun: Function to be decorated
        :return: wrapped :py:obj:`~fun`
        """
        async def inner(msg: AbstractIncomingMessage):
            message = Message.from_dict(**json.loads(msg.body.decode('utf-8')))
            await fun(message)
            await msg.ack()

        return inner

    def add_callback(self, queue: str, callback: Callable[[Message], None], exchange: str = None, is_durable: bool = True):
        """
        Add a callback to be called when a message is received on :py:obj:`~queue`.

        :param queue: Name of queue to consume messages from
        :param callback: Callback function
        :param exchange: Name of exchange to bind queue to
        :param is_durable: Whether to create a durable queue or not
        """
        self.callback_map[queue] = (self.__callback_wrapper(callback), exchange, is_durable)

    async def run(self):
        """
        Run the consumer.

        This will start the consumer, i.e. register configured callbacks under the
        desired queues.
        In case the startup process fails, the consumer will be closed
        and the occurred exception will be raised.

        **Note:** Register callbacks using :py:obj:`~add_callback` before calling this method.
        """
        try:
            await self.connect()
            await self.channel.set_qos(prefetch_count=1)
            for queue_name, queue_definition in self.callback_map.items():
                (callback, exchange_name, is_durable) = queue_definition
                queue = await self.channel.declare_queue(
                    queue_name, durable=is_durable,
                    exclusive=not is_durable
                )
                if exchange_name is not None:
                    exchange = await self.channel.declare_exchange(
                        exchange_name, ExchangeType.FANOUT,
                        durable=True,
                    )
                    await queue.bind(exchange, queue_name)
                await queue.consume(callback, no_ack=False)

            self.logger.info("LISTENING TO {}".format(' '.join(self.callback_map.keys())))
            await asyncio.Future()
        except Exception as e:
            self.logger.exception("Exception in consumer")
            raise e
        finally:
            await self.disconnect()


class ModuleInterface:
    """
    Base class for module implementations.

    :param connection: Connection to RabbitMQ
    :param exchange: Name of exchange to publish messages to
    :param default_consumer_queue: Name of queue to consume messages from

    Example:
        >>> import aio_pika
        >>> import asyncio
        >>> from easybits_module_interface import ModuleInterface
        >>> from easybits_module_interface.models import Message
        >>> class MyModule(ModuleInterface):
        ...     async def callback(self, message: Message):
        ...         print(message)
        >>> connection = await aio_pika.connect("amqp://localhost/", loop=loop)
        >>> module = MyModule(connection, "test_exchange", "test_queue")
        >>> await module.run()
        >>> try:
        >>>     await asyncio.Future()
        >>> finally:
        >>>     await module.shutdown()

    Or as a `system_message` only module:
        >>> class MyModule(ModuleInterface):
        ...     def __init__(self, connection):
        ...         super().__init__(connection, "", None)
        ...         self.add_callback("system_messages_my_module", self.callback)
        ...
        ...     async def callback(self, message: Message):
        ...         print(message)
        ...
        >>> connection = await aio_pika.connect("amqp://localhost/", loop=loop)
        >>> module = MyModule(connection)
    """
    def __init__(
            self, connection, exchange: Optional[str] = None,
            default_consumer_queue: Optional[str] = None
    ) -> None:
        #: Provide placeholder for logger
        self.logger = None

        #: default connection to message broker
        self.__connection = connection

        #: consumer entity to consume incoming messages
        self.__consumer = Consumer(self.__connection)

        if default_consumer_queue is not None:
            #: add default callback for default queue
            self.__consumer.add_callback(default_consumer_queue, self.callback)

        #: publisher entity to publish outgoing messages
        self.__publisher = Publisher(self.__connection, exchange)

        #: initializing the default logger
        self.init_logger('module_interface')

    def __new__(cls, *args, **kwargs):
        # guard against instantiating the base class
        if cls is ModuleInterface:
            raise TypeError(
                f"only children of '{cls.__name__}' may be instantiated"
            )
        return super().__new__(cls)

    def init_logger(self, name: str) -> None:
        """
        Initialize logger for module implementation.

        :param name: Name of logger
        :return: None
        """
        name = f'{name.lower()}.interface'

        self.logger = logging.getLogger(name)
        self.__consumer.set_logger(self.logger)
        self.__publisher.set_logger(self.logger)

    def capture_exception(self, exception) -> None:
        """
        Helper to capture exceptions during message processing.

        Logs exception to:
        - stdout logs
        - sentry

        :param exception: The exception to capture
        :returns: Nothing
        """
        self.logger.error(traceback.format_exc())
        sentry_sdk.capture_exception(exception)

    def get_loggable_message(self, message: Message, exclude_outgoing: bool = False) -> dict:
        """
        Helper method to consistently log messages

        :param message: The message to convert
        :param exclude_outgoing: Indicates if outgoing message should be masked, too
        :returns: masked message dictionary
        """
        msg = asdict(message)
        keys_to_hide = {
            'message.incoming_file': 'FILE_PLACEHOLDER',
        }
        if exclude_outgoing:
            keys_to_hide['message.outgoing'] = 'OUTGOING_PLACEHOLDER'
        for key, placeholder in keys_to_hide.items():
            elems = key.split('.')
            ref = msg
            for idx, e in enumerate(elems):
                can_be_ignored = (
                    ref in [None, False, []]
                    or e not in ref
                    or ref.get(e) in [None, False, []]
                )
                if can_be_ignored:
                    break
                if idx < len(elems) - 1:
                    ref = ref[e]
                else:
                    ref[e] = placeholder
        return msg

    def add_callback(self, queue: str, callback, exchange=None, is_durable=True) -> None:
        """
        Add a callback to be called when a message is received on :py:obj:`~queue`.

        :param queue: Name of queue to consume messages from
        :param callback: Callback function
        :param exchange: Name of exchange to bind queue to
        :param is_durable: Whether to create a durable queue or not
        :return: None
        """
        self.__consumer.add_callback(queue, callback, exchange, is_durable)

    async def run(self) -> None:
        """
        Run the module.

        This will start the consumer and wait for it to finish.
        """
        await self.__publisher.connect()
        await self.__consumer.run()

    async def shutdown(self) -> None:
        """
        Shutdown the module.
        """
        await self.__consumer.disconnect()
        await self.__publisher.disconnect()
        await self.__connection.close()
        self.logger.info("Shutting down")

    async def publish(self, message: Message, exchange: str = None, queue: str = "") -> None:
        """
        Publish a message to RabbitMQ.

        If :py:obj:`~exchange` is empty, the message will be published to the Publisher's default exchange.
        If :py:obj:`~queue` is empty, the message will be published to the RMQ default queue.

        :param message: Message to be published
        :param exchange: Name of exchange to publish message to
        :param queue: Name of queue to publish message to
        :return: None
        """
        await self.__publisher.publish(message, exchange, queue)

    def is_base64(self, s: str) -> bool:
        """
        Check if a string is valid base64.

        :param s: String to check
        :return: True if string is valid base64, False otherwise
        """
        if s is None:
            return False
        try:
            return base64.b64encode(base64.b64decode(s)).decode('utf-8') == s
        except Exception:
            return False

    def get_file_message(self, image_path: str, return_base64: bool = False) -> Optional[tuple[str, str]]:
        """
        Fetch file from URL or base64 data URI.

        :param image_path: URL or data URI of the file
        :param return_base64: If True, return file data as base64 string
        :return: Tuple of (file_data, mime_type) or None if failed
        """
        try:
            if image_path.startswith('data:'):
                # Extract data from data URI
                data_parts = image_path.split(';base64,')
                if len(data_parts) != 2:
                    return None
                mime_type = data_parts[0].split(':')[1]
                image_data = data_parts[1]
                
                # Validate base64
                if not self.is_base64(image_data):
                    return None
                
                # Convert to bytes if not returning base64
                if not return_base64:
                    image_data = base64.b64decode(image_data)
            else:
                # Download file from URL
                response = requests.get(image_path)
                response.raise_for_status()
                image_data = response.content
                
                # Determine mime type
                if '.png' in image_path:
                    mime_type = 'image/png'
                elif '.jpg' in image_path or '.jpeg' in image_path:
                    mime_type = 'image/jpeg'
                else:
                    mime_type = mimetypes.guess_type(image_path.split('?')[0])[0]
                
                # Convert to base64 if requested
                if return_base64:
                    image_data = base64.b64encode(image_data).decode('utf-8')

            if not mime_type or not image_data:
                return None
                
            return (image_data, mime_type)
        except Exception:
            return None

    async def callback(self, message: Message) -> None:
        """
        Callback function to be implemented by module implementations.

        :param message: Message received from consumer queue
        :return: None
        """
        raise NotImplementedError("Callback not implemented")



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from easybits_module_interface.models.base import DynamicFieldsDataClass, BaseDataClass
from easybits_module_interface.models.module import AvailableModule, BotModule, ContextProvider


@dataclass
class Tool(BaseDataClass):
    id: int
    position: int
    template: str
    send_intermediate_result: bool = False
    status_report: Optional[str] = None


@dataclass
class RouterConfig(BaseDataClass):
    tool_sequence: dict[str, Tool]

    @classmethod
    def from_dict(cls, **kwargs) -> 'RouterConfig':
        """
        Classmethod to create a RouterConfig instance from a dictionary

        :param kwargs: Dictionary containing the RouterConfig data
        :returns: A freshly created RouterConfig instance
        """
        tool_sequence = kwargs.pop('tool_sequence', [])
        
        return super().from_dict(
            **kwargs,
            tool_sequence={str(val.get('position', idx)): Tool.from_dict(**val) for idx, val in enumerate(tool_sequence)}
        )


@dataclass
class ToolChainParserConfig(BaseDataClass):
    router_config: RouterConfig | None = None
    purpose: str | None = None
    command: str | None = None

    @classmethod
    def from_dict(cls, **kwargs) -> 'ToolChainParserConfig':
        """
        Classmethod to create a ToolChainParserConfig instance from a dictionary

        :param kwargs: Dictionary containing the ToolChainParserConfig data
        :returns: A freshly created ToolChainParserConfig instance
        """
        router_config = kwargs.pop('router_config', [])
        return super().from_dict(
            **kwargs,
            router_config=RouterConfig.from_dict(**router_config) if router_config else None
        )

    def __getitem__(self, name):
        return getattr(self, name)

    def get(self, name, default=None):
        return getattr(self, name, default)


@dataclass
class ToolChainModule(BaseDataClass):
    """
    ToolChainModule entity to represent a BotModule that is part of a ToolChain
    """
    id: int
    module: AvailableModule
    parser_configuration: dict = field(default_factory=dict)
    context_providers: list[ContextProvider] = field(default_factory=list)
    configuration_values: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, **kwargs) -> 'ToolChainModule':
        """
        Classmethod to create a ToolChainModule instance from a dictionary

        :param kwargs: Dictionary containing the ToolChainModule data
        :returns: A freshly created ToolChainModule instance
        """
        context_providers = kwargs.pop('context_providers', [])
        module = kwargs.pop('module', {})

        return super().from_dict(
            **kwargs,
            module=AvailableModule.from_dict(**module) if module else None,
            context_providers=[ContextProvider.from_dict(**c) for c in context_providers]
        )


@dataclass
class ToolChain(BaseDataClass):
    """
    ToolChain entity to represent a sequence of BotModules that are executed in a sequence
    """
    id: int
    name: str
    tool_chain_modules: list[ToolChainModule]
    is_active: bool = True
    parser_configuration: ToolChainParserConfig | None = None
    has_reengagement: bool = False
    reengagement_message: str | None = None
    reengagement_timeout_minutes: int | None = None

    @property
    def internal_id(self) -> str:
        return f'T{self.id}'

    @property
    def intentions(self) -> List[str]:
        """
        Getter for the intentions of the ToolChain

        :returns: List of intentions
        """
        return self.get_module_by_index(0).module.intentions

    def get_module_by_index(self, index):
        """
        Getter for a ToolChainModule by index

        :param index: Index of the ToolChainModule to retrieve
        :returns: ToolChainModule instance
        """
        tool = self.parser_configuration.router_config.tool_sequence.get(str(index))
        if not tool:
            return None
        module = [chain_module for chain_module in self.tool_chain_modules if chain_module.id == tool.id]
        return module[0] if module else None

    @property
    def module(self) -> AvailableModule:
        return self.get_module_by_index(0).module

    @property
    def configuration_values(self) -> dict:
        return self.get_module_by_index(0).configuration_values

    @property
    def context_providers(self) -> List[ContextProvider]:
        return self.get_module_by_index(0).context_providers

    @classmethod
    def from_dict(cls, **kwargs) -> 'ToolChain':
        """
        Classmethod to create a ToolChain instance from a dictionary

        :param kwargs: Dictionary containing the ToolChain data
        :returns: A freshly created ToolChain instance
        """
        tool_chain_modules = kwargs.pop('tool_chain_modules', [])
        config = kwargs.pop('parser_configuration', {})
        return super().from_dict(
            **kwargs,
            parser_configuration=ToolChainParserConfig.from_dict(**config) if config else None,
            tool_chain_modules=[ToolChainModule.from_dict(**i) for i in tool_chain_modules]
        )
